""" Parser module """ 

import h5py

def get_patch_list_AVBP(mesh_file):
    """
    *Get the list of patches of a mesh hdf5 file*
    
    :param mesh_file: Path to mesh file
    :type mesh_file: str
    
    :return:
        
        - **patch_list** - List of patches
    """
    with h5py.File(mesh_file, 'r') as fin:
        patch_list = fin['Boundary/PatchLabels'][()]
    patch_list = [patch.decode('UTF-8').strip() for patch in patch_list]
    return patch_list

def parse_meshinfo(msgs):
    """
    *Parse HIP output to get mesh informations*
    :param msgs: List of string message line

    :returns:

        - **meshinfo** - (str) Mesh informations
    """
    in_found = False
    for line in msgs:
        if 'read' in line or 'generate' in line:
            idx_in = msgs.index(line) + 1
            in_found = True
        elif 'copy 3D' in line:
            idx_in = msgs.index(line) + 1
            in_found = True
        elif 'copy q2t' in line or 'copy 2tets' in line:
            idx_in = msgs.index(line) + 1
            in_found = True
        elif in_found and ' <<hip[' in line:
            idx_out = msgs.index(line)
            in_found = False

    meshinfo = "\n".join(map(str.strip, msgs[idx_in:idx_out]))
    return meshinfo

def parse_versioninfo(msgs):
    """
    *Parse HIP output to get hip version informations*
    :param msgs: List of string message line

    :returns:

        - **meshinfo** - (str) Mesh informations
    """
    line = msgs[1].split(',')
    version = line[1][9:]
    version = "Version : " + "".join(version)[:-1]
    date = line[2].strip()

    version_info = "\n".join([version, date]) + '\n'
    return version_info

def parse_boundaries(msgs):
    """
    """
    patch_dict = {}
    found = False
    for idx, line in enumerate(msgs):
        if 'list surface' in line:
            found = True
            break
    if not found:
        return patch_dict

    idx += 2
    for line in msgs[idx:]:
        if line.strip() == '':
            break
        elements = [element.strip() for element in line.split(',')]
        idx += 1
        patch = elements[-1]
        patch_dict[patch] = {
            'nbr': elements[0].split(':')[0],
            'type': elements[2],
            'order': elements[3]}

    return patch_dict

def parse_periodicities(msgs):
    """
    *Parse HIP output to get peridic patches*
    :param msgs: List of string message line

    :returns:

        - **perio_list** - List of periodic patch dict['name', \
                                                       'leader_patch', \
                                                       'follower_patch']
    """
    perio_list = []
    found = False
    for idx, line in enumerate(msgs):
        if 'list periodic' in line:
            found = True
            break
    if not found:
        return perio_list

    idx += 1
    block = 9
    id_ = 'id'
    while '<<hip[' not in msgs[idx]:
        patches = msgs[idx].split()
        perio_patch = {'name': id_,
                       'leader_patch': patches[1],
                       'follower_patch': patches[0]}
        perio_list.append(perio_patch)
        idx += block
        id_ += '#'

    return perio_list



"""Module containing hip readers functions """
import os
from pyhip.hipster import pyhip_cmd


__all__ = ["read_mesh_files",
           "read_centaur_mesh",
           "read_cgns_mesh",
           "read_ensight_mesh",
           "read_fluent_mesh",
           "read_gmsh_mesh",
           "read_hdf5_mesh"]

def _split_file_path(file_path):
    """ Returns the file basename and the parent
        directory given file path

        Parameters:
        ==========
        file_path : path to file (absolute or relative)

        Returns:
        =======
        parent : file parent directory
        filename : file name
    """
    abs_path = os.path.abspath(file_path)
    parent = os.path.dirname(abs_path)
    filename = os.path.basename(abs_path)
    return parent, filename

def _rel_path(files, start_path):

    abspaths = [os.path.abspath(file) for file in files]
    rel_paths = [os.path.relpath(abspath, start=start_path)
                 for abspath in abspaths]
    return rel_paths

def read_mesh_files(mesh_files, mesh_format, **kwargs):
    """ read a mesh file
        Parameters:
        ==========
        mesh_file : A list of mesh files paths (absol or rel)
        mesh_format : Format of the mesh file (e.g hdf5)

        Returns:
        =======
        None
    """
    files = [file for file in mesh_files if file is not None]
    parent_dir, _ = _split_file_path(files[0])
    files = _rel_path(files, parent_dir)
    if mesh_format.strip().lower() == 'hdf5' and len(files) == 2:
        files = "-a %s -s %s" %(files[0], files[1])
    else:
        files = ' '.join(files)

    commands = ["set path %s"%parent_dir]
    commands.append("read %s %s" %(mesh_format, files))
    commands.append("var")
    for command in commands:
        _ = pyhip_cmd(command, **kwargs)
    return commands

# def read_bae_mesh(fro_file, gri_file, bco_file):
#     """Read Britsh Aerospace and Swansea University
#         tetrahedral meshes.
#def _rel_path(files, start_path):

#         Parameters:
#         ==========
#         fro_file : boundary faces and singular lines file (binary)
#         gri_file : connectivity and coordinates file (binary)
#         bco_file : file containing boundary condition to the
#                   various surfaces mapping. (ascii)

#     """
#     return read_mesh_files([fro_file, gri_file, bco_file], "bae")

# def read_cedre_mesh(cdre_mesh_file):
#     """Read unstructured face-based mesh format used Cedre code.

#        Parameters:
#        ==========
#        cdre_mesh_file : Cedre unstructured face-based mesh file
#     """
#     return read_mesh_files([cdre_mesh_file], "cdre")

def read_centaur_mesh(centaur_mesh_file, **kwargs):
    """Read hybrid grid in Centaursofts format.

       Parameters:
       ==========
       centaur_mesh_file : Centaursofts format mesh_file. Supported file
                           format are versions:
                           - 4 (single record)
                           - 5 (multiple record).
    """
    return read_mesh_files([centaur_mesh_file], "centaur", **kwargs)

def read_cgns_mesh(grid_file, sol_file=None, abnd_file=None, **kwargs):
    """Read an unstructured CGNS database.

       Parameters:
       ==========
       grid_file: CGNS grid file
       sol_file: CGNS solution file
       abnd_file: if bnd info in an AVBP-style .asciiBound file is given,
                  this supersedes boundary condition definition in
                  the CGNS grid_file.

    """
    return read_mesh_files([grid_file, sol_file, abnd_file], "cgns", **kwargs)

def read_ensight_mesh(ensight_case_file, **kwargs):
    """Read an unstructured grid (no Solution) in
       Ensight Gold format

       Parameters:
       ==========
       ensight_case_file: Ensight Gold format file

    """
    return read_mesh_files([ensight_case_file], "ensight", **kwargs)

def read_fluent_mesh(mesh_file, solution_file=None, **kwargs):
    """Read an unstructured grid and optionally
       a solution in Fluent’s v5 and v6 ascii and binary
       formats

       Parameters:
       ==========
       mesh_file: Fluent/Gambit mesh file (.msh/.cas extensions)
       solution_file : Fluent solution file (.dat extension)
    """
    return read_mesh_files([mesh_file, solution_file], "fluent", **kwargs)

def read_gmsh_mesh(gmsh_mesh_file, list_variables=None, **kwargs):
    """Read an unstructured grid in gmsh ASCII format

       Parameters:
       ==========
       gmsh_mesh_file: gmsh mesh file
       list_variables : a list of variable files
                        (Up to MAX UNKNOWNS = 256)
    """
    if list_variables is None:
        cmds = read_mesh_files([gmsh_mesh_file], "gmsh", **kwargs)
    else:
        cmds = read_mesh_files([gmsh_mesh_file, *list_variables], "gmsh", **kwargs)
    return cmds

def read_hdf5_mesh(grid_file, sol_file=None, **kwargs):
    """Read an unstructured grid in hdf5 format

       Parameters:
       ==========
       grid_file: hdf5 mesh file
       sol_file : Solution_file
    """
    return read_mesh_files([grid_file, sol_file], "hdf5", **kwargs)



cimport cython

cdef extern from "../../../src/hip.c":
    DEF TEXT_LEN = 1024
    DEF MAX_DIM = 3
    DEF IPTR64 = 1
    DEF INT_MAX = 2147483647
    DEF MAX_CPT_CHUNKS = INT_MAX
    DEF MAX_CPT_VXNR = INT_MAX
    DEF MAX_BC_CHAR = 81
    DEF HIP_USE_ULONG = 1
    DEF MAX_ZONE_FIELD_LEN = 8 
    DEF MAX_EDGES_ELEM = 12
    DEF MAX_COLORS = 255
    DEF MAX_DRV_ELEM_TYPES = 29
    DEF MAX_ELEM_TYPES = 6
    DEF MAX_VX_FACE =4
    DEF MAX_EDGES_FACE = 4
    DEF MAX_CHILDS_FACE = 8
    DEF MAX_VX_ELEM = 8
    DEF MAX_FACES_ELEM = 6
    DEF MAX_CHILDS_ELEM = 10
    DEF LEN_GRPNAME = 30
    DEF MAX_UNKNOWNS = 256
    DEF LEN_VARNAME = 30

cdef extern from "../../../src/hip.c":
    void c_init "hip_init"()

    ctypedef enum hip_stat_e:
        success, fatal, warning, info, blank


    ctypedef enum bcGeoType_e:
        bnd, match, inter, duplicateInter, cut,
        noBcGeoType, bndAndInter, any

    ctypedef enum spec_bc_e:
        unspecBc, perBc, wallBc, otherBc

    ctypedef enum specialTopo_e:
        noTopo, axiX, axiY, axiZ, noBc, surf

    ctypedef enum grid_type_enum:
        noGr, mb, uns

    ctypedef enum varCat_e:
        noCat, ns, species, rrates, tpf, rans,
        add, mean, fictive, add_tpf, param, other, hip

    ctypedef struct var_s:
        varCat_e cat
        char grp[LEN_GRPNAME]
        char name[LEN_VARNAME]
        int isVec
        int flag

    ctypedef enum varType_e:
        noVar, cons, prim, primT, para, noType 


    ctypedef struct varList_s:
        int mUnknowns 
        int mUnknFlow
        varType_e varType
        var_s var[MAX_UNKNOWNS]
        double freeStreamVar[MAX_UNKNOWNS]

    ctypedef struct bc_struct:
        char text[MAX_BC_CHAR]
        spec_bc_e bc_e
        char type[MAX_BC_CHAR]
        int nr
        int order
        int refCount
        int mark
        bc_struct *PprvBc
        bc_struct *PnxtBc
        bcGeoType_e geoType
        double llBox[MAX_DIM]
        double urBox[MAX_DIM]
        double rllBox[2]
        double rurBox[2] 


    ctypedef struct block_struct:
        int nr
        char name[TEXT_LEN]
        mb_struct *PmbRoot
        int mVertFile[MAX_DIM]
        int mVert[MAX_DIM]
        int skip
        int mVertsBlock
        double *Pcoor
        int mVertsMarked
        double *PdblMark
        int *PintMark
        double *Punknown
        int mElemsBlock
        int mElemsMarked
        int *PelemMark
        int mSubFaces
        double llBox[MAX_DIM]
        double urBox[MAX_DIM]
        double hMin
        double hMax

    ctypedef struct mb_struct:
        int mBlocks
        block_struct *PblockS
        int mDim
        int mUnknowns

        int mVerts
        int mElems

        int mSubFaces
        varList_s varList


    ctypedef struct color_s:
        unsigned int mBnd
        unsigned int color
        unsigned int mark
        unsigned int mVxColl
        unsigned int maxColl

    ctypedef struct uns_s:
        int nr
        char name[TEXT_LEN]
        grid_struct *pGrid

        int validGrid
        int adapted
        specialTopo_e specialTopo
        int mDim
        double llBox[MAX_DIM]
        double urBox[MAX_DIM]
        double llBoxCyl[2]
        double urBoxCyl[2]
        double hMin
        double hMax
        double volElemMin
        double volElemMax
        double volDomain
        double epsOverlap
        double epsOverlapSq
        color_s *pVxColor
        varList_s varList
        int mBc
        int mBcBnd
        bc_struct **ppBc
        uns_s *pUnsFine
        uns_s *pUnsCoarse
        uns_s *pUnsFinest

    ctypedef struct uns_c:
        grid_struct *PnxtGrid
        grid_struct *PprvGrid
        int nr
        int mDim
        char text[TEXT_LEN]
        varList_s *pVarList
        grid_type_enum type
        uns_s *pUns


    ctypedef struct mb_c:
        grid_struct *PnxtGrid
        grid_struct *PprvGrid
        int nr
        int mDim
        char text[TEXT_LEN]
        varList_s *pVarList
        grid_type_enum type
        mb_struct *Pmb

    ctypedef union grid_struct:
        mb_c mb
        uns_c uns

    ctypedef struct ret_s:
        grid_struct *pGrid
        uns_s *pUns
        hip_stat_e status
        char *msg

    ret_s _call_hip_menus "hip_menu"( char *Pstring )

cpdef hip_init():
    c_init()

cpdef hip_cmd(input_):
    input_bytes = bytes(input_, 'utf-8')
    ret = _call_hip_menus(input_bytes)

    msg = ret.msg.decode("utf-8")
    status = int(ret.status)
    return status, msg
    ### Example for further improvement
    ### To access a member of a structure 
    ### It is done within the same format
    ### As for python objects, for example
    ### Accessing "name" member of "pUns":
        # name = ret.pUns.name.decode("utf-8")
    ### Other examples for non string
        # varlist = ret.pUns.varList
        # nr = int(ret.pUns.nr)
        # mDim = int(ret.pUns.mDim)
        # print(type(varlist))





from tiny_3d_engine import load_file_as_scene
from nob.core import Tree
from opentea.process_utils import update_3d_callback
import os

def update_3d_scene(nob_in, scene):
    """Update the list of dimensions."""

    nob_tree = Tree(nob_in)
    geo_path = nob_tree.inputs.geofile[:]
    try:
        scene = load_file_as_scene(geo_path, prefix='in', scene=scene, color="#ff0000")
        os.remove(geo_path)
    except KeyError:
        print("Scene couldn't be loaded. Might be invalid .geo file.\n")
    except:
        print("Scene couldn't be loaded.\n")

    scene.add_axes()

    return scene


if __name__ == "__main__":
    print("3D CBck:")
    update_3d_callback(update_3d_scene)



"""
*Module of generation of a mesh ID Card*

This module generate a mesh ID Card summing up main mesh
informations and displaying geometry.

"""
from io import BytesIO
from collections import namedtuple
import h5py
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib.ticker import PercentFormatter
from reportlab.pdfgen import canvas
from reportlab.lib.units import cm, inch
from reportlab.lib.utils import ImageReader
from scipy.spatial.transform import Rotation as Rot
from arnica.utils.vector_actions import (
    renormalize,
    rotate_vect_around_axis,
)

__all__ = ['generate_mesh_idcard']

MARGIN = 1.*cm
THOUSAND_SEP = "'"
LINE_HEIGHT = 0.35*cm
Bounds = namedtuple('Bounds', ['min', 'max'])
mid_ = lambda arr: (arr.max(axis=0) + arr.min(axis=0)) / 2.
range_ = lambda arr: arr.max(axis=0) - arr.min(axis=0)


def _get_datasets(fin, ds_names, data_dict=None, path_='', gr_skip=None):
    """ Get list of dataset from hdf file """
    if data_dict is None:
        data_dict = dict()
    if gr_skip is None:
        gr_skip = []
    for gname, group in fin.items():
        path = path_ + "/" + gname
        if not isinstance(group, h5py.Dataset) and gname not in gr_skip:
            _get_datasets(fin[gname], ds_names, data_dict, path_=path, gr_skip=gr_skip)
        else:
            if gname in ds_names:
                data_dict[gname] = fin[path][()]
    return data_dict

def extract_hdf_meshinfo(meshfile):
    #pylint: disable=too-many-locals
    """
    *Get mesh informations from hdf file and format them*

    :param meshfile: Path to hdf mesh file
    :type meshfile: str

    :returns:

        - **meshinfo** - Dict of mesh informations
        - **idx_patches** - List of patch index (starting by 1) without periodic patches
    """
    ds_names = (
        'hipversion',
        'vol_domain',
        'h_min',
        'h_max',
        'r_min',
        'r_max',
        'vol_elem_max',
        'vol_elem_min',
        'x_min',
        'x_max',
        'bnode_lidx',
        'periodic_patch',
        'periodic_angle',
        'PatchLabels')
    connectivities = {
        'tri->node': {'name': 'Triangles', 'nnode': 3},
        'qua->node': {'name': 'Quadrilateral', 'nnode': 4},
        'tet->node': {'name': 'Tetrahedra', 'nnode': 4},
        'pyr->node': {'name': 'Pyramids', 'nnode': 5},
        'pri->node': {'name': 'Prisms', 'nnode': 6},
        'hex->node': {'name': 'Hexahedra', 'nnode': 8}}

    with h5py.File(meshfile, 'r') as fin:
        data_dict = _get_datasets(fin, ds_names)
        connectivities_dict = _get_datasets(fin, connectivities.keys())

    if 'periodic_angle' in data_dict:
        is_axi = True
    else:
        is_axi = False
    ndim = len(data_dict['x_min'])

    meshinfo = {}
    meshinfo['Mesh name'] = meshfile.split('/')[-1]
    if isinstance(data_dict['hipversion'], np.ndarray):
        data_dict['hipversion'] = data_dict['hipversion'][0]
    meshinfo['Hip version'] = data_dict['hipversion'].decode('utf-8')
    meshinfo['Number of nodes'] = 0
    meshinfo['Number of boundary nodes'] = int(data_dict['bnode_lidx'][-1])
    meshinfo['Domain volume [m3]'] = data_dict['vol_domain'][0]
    meshinfo['Number of cells'] = {}

    for c_type, dataset in connectivities_dict.items():
        name = connectivities[c_type]['name']
        nnode = connectivities[c_type]['nnode']
        ncells = dataset.size / nnode
        meshinfo['Number of cells'][name] = int(ncells)
        meshinfo['Number of nodes'] += int(dataset.max())

    meshinfo['Metric'] = {}
    meshinfo['Metric']['Edge length [m]'] = Bounds(
        data_dict['h_min'][0],
        data_dict['h_max'][0])
    meshinfo['Metric']['Element volume [m3]'] = Bounds(
        data_dict['vol_elem_min'][0],
        data_dict['vol_elem_max'][0])

    meshinfo['Bounding box'] = {}
    meshinfo['Bounding box']['x [m]'] = Bounds(
        data_dict['x_min'][0],
        data_dict['x_max'][0])
    if ndim == 2:
        meshinfo['Bounding box']['y [m]'] = Bounds(
            data_dict['x_min'][1],
            data_dict['x_max'][1])
    else:
        if is_axi:
            meshinfo['Bounding box']['r [m]'] = Bounds(
                data_dict['r_min'][0],
                data_dict['r_max'][0])
            meshinfo['Bounding box']['theta [deg]'] = Bounds(
                data_dict['r_min'][1],
                data_dict['r_max'][1])
        else:
            meshinfo['Bounding box']['y [m]'] = Bounds(
                data_dict['x_min'][1],
                data_dict['x_max'][1])
            meshinfo['Bounding box']['z [m]'] = Bounds(
                data_dict['x_min'][2],
                data_dict['x_max'][2])

    patch_labels = np.char.strip(np.char.decode(data_dict['PatchLabels']))
    if 'periodic_angle' in data_dict:
        meshinfo['Periodic angle [deg]'] = round(data_dict['periodic_angle'][0], 5)
    if 'periodic_patch' in data_dict:
        idx_perio = data_dict['periodic_patch'] - 1
        idx_perio = idx_perio.reshape((int(idx_perio.size/2), 2))
        meshinfo['Periodic pairs'] = {f"{i*' '}": f"{pair[0]} - {pair[1]}"
                                      for i, pair in enumerate(patch_labels[idx_perio])}

    return meshinfo, is_axi, ndim, patch_labels

def arrange_meshinfo(meshinfo_dict, col_width=None, lines=None, shift=0):
    """
    *Format mesh informations as lines to be drawn on pdf*

    :param meshinfo_dict: Dict of mesh informations

    :optional arguments:

    :param col_width: Width in digit of label column
    :type col_width: int
    :param lines: List of lines to be drawn
    :param shift: Tabulation shift in digit for sub elements
    :type shift: int
    """
    if lines is None:
        lines = []
    if col_width is None:
        col_width = max([len(label) for label in meshinfo_dict.keys()]) + 3

    for label, info in meshinfo_dict.items():
        if isinstance(info, dict):
            lines.append(f"{' '*shift}{label:{col_width}}")
            arrange_meshinfo(info, col_width=col_width, lines=lines, shift=shift+2)
        else:
            if isinstance(info, Bounds):
                if abs(info.min) > 1.e-4 and abs(info.max) > 1.e-4:
                    value = f"min={f'{info.min:6f}':17} max={info.max:6f}"
                else:
                    value = f"min={f'{info.min:6e}':17} max={info.max:6e}"
            elif isinstance(info, int):
                value = f"{info:,}".replace(',', THOUSAND_SEP)
            elif isinstance(info, str):
                value = info
            else:
                if abs(info) > 1.e-4:
                    value = f"{info:6f}"
                else:
                    value = f"{info:6e}"
            lines.append(f"{' '*shift}{label:{col_width-shift}}" + value)

    return lines

def _create_perspective_mark(xyz_rough, *rot_tuples, coef=100.):
    """
    *Generate a perspective xyz array*

    :param xyz_rough: Array of dim (n,3) of xyz-coordinates
    :param rot_tuples: List of tuple of rotation data (Axis, angle)
    :param coef: Coefficient of perspective
    :type coef: float

    :returns: Array of dim (n,3) of perspectived xyz-coordinates
    """
    center = (xyz_rough.min(axis=0) + xyz_rough.max(axis=0)) / 2.
    xyz_rough -= center
    xyz_persp = np.multiply(xyz_rough,
                            np.power(coef, (-xyz_rough[:, 0][:, np.newaxis])))
    xyz_persp = rotate_vect_around_axis(xyz_persp, *rot_tuples)
    return xyz_persp

def _get_boundary_nodes(meshfile, ndim):
    """ Return boundary nodes array """
    axis = ['x', 'y', 'z'][:ndim]
    coords = [[] for _ in range(ndim)]

    with h5py.File(meshfile, 'r') as fin:
        for idx in fin[f'Patch/'].keys():
            for i, axes in enumerate(axis):
                coords[i].extend(fin[f'Patch/{idx}/Coordinates/{axes}'][()])
    return np.asarray(coords)

def _plot_skin(axis, coords, labels):
    """Plot hist2d from coords"""

    def get_bins(x_crd, y_crd):
        """Computes equal bins from aspect ratio"""
        ratio = range_(y_crd) / range_(x_crd)
        return int(250.), int(250. * ratio)

    plt.rcParams['image.cmap'] = 'Blues'
    axis.hist2d(*coords,
                bins=get_bins(*coords),
                norm=LogNorm())
    axis.set_aspect('equal')
    axis.set_xlabel(labels[0])
    axis.set_ylabel(labels[1])
    axis.xaxis.set_ticks([coords[0].min(), coords[0].max()])
    axis.yaxis.set_ticks([coords[1].min(), coords[1].max()])
    axis.yaxis.set_label_coords(-0.1, 0.5)
    for edge in ['top', 'bottom', 'right', 'left']:
        axis.spines[edge].set_visible(False)

def plot_2d_density_view(meshfile): 
    """
    *Plot front 2D mesh view as hist2d*

    :param meshfile: Path to hdf5 mesh file
    :type meshfile: str
    """
    x_arr, y_arr = _get_boundary_nodes(meshfile, 2)

    fig_width = 10.
    aspect_ratio = range_(y_arr) / range_(x_arr)
    fig, axis = plt.subplots(figsize=(fig_width, fig_width * aspect_ratio))
    _plot_skin(axis, (x_arr, y_arr), ('x', 'y'))
    fig.subplots_adjust(left=0.125, right=0.95, bottom=0.1, top=0.95)

    buffer_hist2d = BytesIO()
    fig.savefig(buffer_hist2d, format='png', transparent=True, dpi=600)
    buffer_hist2d.seek(0)
    return buffer_hist2d

def plot_3d_density_view(meshfile, is_axi):
    #pylint: disable=too-many-locals
    """
    *Plot 3 sides and 1 perspective 3D mesh views as hist2d*

    :param meshfile: Path to hdf5 mesh file
    :type meshfile: str
    :param is_axi: Determine if geometry is axi-periodic.
    :type is_axi: bool
    """
    x_arr, y_arr, z_arr = _get_boundary_nodes(meshfile, 3)
    x_per, y_per, _ = _create_perspective_mark(
        np.stack((x_arr, y_arr, z_arr), axis=-1),
        ([0, 1, 0], 60.),
        ([0, 0, 1], 10.)).T
    aspect_ratio = min(range_(z_arr) / range_(x_per),
                       range_(z_arr) / range_(y_per))
    x_per *= aspect_ratio
    x_per += mid_(z_arr) - mid_(x_per)
    y_per *= aspect_ratio
    y_per += mid_(z_arr) - mid_(y_per)

    Views = namedtuple('View', ['labels', 'coords'])
    if is_axi:
        r_arr = np.hypot(y_arr, z_arr)
        view_axes = [Views(('x', 'r'), (x_arr, r_arr)),
                     Views(('z', 'y'), (z_arr, y_arr)),
                     Views(('x', 'z'), (x_arr, z_arr))]
    else:
        view_axes = [Views(('x', 'y'), (x_arr, y_arr)),
                     Views(('z', 'y'), (z_arr, y_arr)),
                     Views(('x', 'z'), (x_arr, z_arr))]

    plt.rcParams['image.cmap'] = 'Blues'
    fig_width = 10.
    aspect_ratio = (range_(y_arr) + range_(z_arr)) / (range_(x_arr) + range_(z_arr))
    fig, axes = plt.subplots(
        figsize=(fig_width, fig_width * aspect_ratio),
        nrows=2, ncols=2,
        sharex='col', sharey='row',
        gridspec_kw={'height_ratios': (range_(y_arr), range_(z_arr)),
                     'width_ratios': (range_(x_arr), range_(z_arr)),
                     'hspace': 0.12,
                     'wspace': 0.12})

    _plot_skin(axes[1, 1], (x_per, y_per), ('', ''))
    for i, view in enumerate(view_axes):
        _plot_skin(axes[i//2, i%2], view.coords, view.labels)
    fig.align_ylabels(axes[:, 0])
    fig.subplots_adjust(left=0.125, right=0.95, bottom=0.1, top=0.95)

    buffer_hist2d = BytesIO()
    fig.savefig(buffer_hist2d, format='png', transparent=True, dpi=600)
    buffer_hist2d.seek(0)
    return buffer_hist2d

def plot_hist_from_meshfile(meshfile, width, height):
    """
    *Plot histogram of dataset occurence*

    :param meshfile: Path to hdf5 mesh file
    :type meshfile: str

    :returns: **buffer_hist** - BytesIO object of the plot
    """
    with h5py.File(meshfile, 'r') as fin:
        vol_node = _get_datasets(fin, ['volume'])['volume']

    vol_cell = vol_node / 4.5
    len_edge = np.power(vol_cell * 12. / np.sqrt(2.), 1. / 3.)

    bins = np.logspace(np.log10(len_edge.min()),
                       np.log10(len_edge.max()),
                       100)
    plt.figure(figsize=(width / inch, height / inch))
    plt.hist(len_edge,
             bins=bins,
             weights=np.ones_like(len_edge)/len_edge.size)
    plt.axvline(len_edge.min(), color='k', linestyle='dashed', linewidth=1)
    plt.axvline(len_edge.max(), color='k', linestyle='dashed', linewidth=1)
    _, max_ylim = plt.ylim()
    plt.text(
        len_edge.min(),
        0.85*max_ylim,
        f' Min = {len_edge.min():.3e} ',
        horizontalalignment='left')
    plt.text(
        len_edge.max(),
        0.85*max_ylim,
        f' Max = {len_edge.max():.3e} ',
        horizontalalignment='right')
    #plt.grid()
    plt.xlabel('Edge length [m]')
    plt.xscale('log')
    plt.gca().yaxis.set_major_formatter(PercentFormatter(decimals=2))
    plt.title('Histogram of Edge length')
    plt.tight_layout()

    buffer_hist = BytesIO()
    plt.savefig(buffer_hist, format='png', transparent=True, dpi=400)
    buffer_hist.seek(0)
    return buffer_hist

def _get_plot_size(width_max, height_max, width_, height_):
    """ Compute fig size to fit in dimensions max available """
    ratio_width = width_max / width_
    ratio_height = height_max / height_
    ratio = min(ratio_width, ratio_height) * 0.99
    return (width_*ratio, height_*ratio)

def _make_title(pdf, mesh_name, coords, width, height):
    """ Draw title section """
    offset = 0.1*cm
    section_height = 1.5*cm
    title = mesh_name.upper() + ' MESH CARD'
    coords[1] -= section_height

    pdf.rect(
        coords[0] + offset,
        coords[1] + offset,
        width - 2 * (MARGIN + offset),
        section_height - 2 * offset)
    pdf.setFont('Courier', 20)
    pdf.drawCentredString(
        width / 2.,
        height - MARGIN - section_height / 2. - 6.,
        title)

def _make_meshinfo(pdf, meshinfo_dict, coords):
    """ Draw meshinfo section """
    pdf.setFont('Courier', 8)
    coords[0] += 0.3*cm
    coords[1] -= LINE_HEIGHT * 1.3
    meshinfo_lines = arrange_meshinfo(meshinfo_dict)

    for line in meshinfo_lines:
        pdf.drawString(*coords, line)
        coords[1] -= LINE_HEIGHT

    coords[0] -= 0.3*cm

def _make_hist(pdf, meshfile, coords, size_max):
    """ Draw histogram section """
    size_max = [side - 0.2*cm for side in size_max]
    buffer = plot_hist_from_meshfile(meshfile, *size_max)
    hist = ImageReader(buffer)
    size = _get_plot_size(*size_max, *hist.getSize())
    coords[1] -= size[1] + 0.1*cm
    pdf.drawImage(hist, *coords, *size)


def _make_meshview(pdf, meshfile, is_axi, ndim, coords, size_max, mid=False):
    #pylint: disable=too-many-arguments
    """ Draw meshview section """
    size_max = [side - 0.2*cm for side in size_max]

    if ndim == 2:
        buffer_plots = plot_2d_density_view(meshfile)
    else:
        buffer_plots = plot_3d_density_view(meshfile, is_axi)
    plot = ImageReader(buffer_plots)
    size = _get_plot_size(*size_max, *plot.getSize())
    coords[1] -= size[1] + 0.1*cm
    if mid is True:
        coords[0] += (size_max[0] - size[0]) / 2.
        coords[1] -= (size_max[1] - size[1]) / 2.
    pdf.drawImage(plot, *coords, *size)
    coords[1] -= 0.1*cm

def generate_landscape(meshfile, pagesize):
    """
    *Generate mesh ID cart from mesh in landscape format*

        - Get mesh informations
        - Build title
        - Build mesh info section
        - Build plots section
        - Build volume vertex histogram

    :param meshfile: Path to hdf5 mesh file
    :type meshfile: str
    :param pagesize: Tuple (width, height) of the page size. \
                     1 pt = 1/72 inch ~ 0.035277 cm
    """
    meshinfo_dict, is_axi, ndim, _ = extract_hdf_meshinfo(meshfile)

    pdf_file = meshinfo_dict['Mesh name'].split('.')[0] + '.mesh.pdf'
    pdf = canvas.Canvas(pdf_file, pagesize=pagesize, verbosity=0)
    width, height = pagesize

    # Title section
    coords = [MARGIN, height - MARGIN]
    mesh_name = meshinfo_dict['Mesh name'].split('.')[0]
    _make_title(pdf, mesh_name, coords, *pagesize)

    pdf.line(*coords, width - MARGIN, coords[1])

    # Meshinfo section
    meshinfo_coords = coords.copy()
    _make_meshinfo(pdf, meshinfo_dict, coords)

    pdf.line(*coords, 0.4*width, coords[1])

    # Histogram section
    size_max = [0.4*width - MARGIN, coords[1] - MARGIN]
    _make_hist(pdf, meshfile, coords, size_max)

    # MeshView section
    coords = [0.4 * width, meshinfo_coords[1]]
    size_max = [0.6 * width - MARGIN, meshinfo_coords[1] - MARGIN]
    _make_meshview(pdf, meshfile, is_axi, ndim, coords, size_max, mid=True)

    pdf.line(0.4*width, meshinfo_coords[1], 0.4*width, MARGIN)

    # Main rectangle
    pdf.rect(MARGIN, MARGIN, width - 2*MARGIN, height - 2*MARGIN)

    print(f'{pdf_file} file saved.')
    pdf.save()


def generate_portrait(meshfile, pagesize):
    """
    *Generate mesh ID cart from mesh in portrait format*

        - Get mesh informations
        - Build title
        - Build mesh info section
        - Build plots section
        - Build volume vertex histogram

    :param meshfile: Path to hdf5 mesh file
    :type meshfile: str
    :param pagesize: Tuple (width, height) of the page size. \
                     1 pt = 1/72 inch ~ 0.035277 cm
    """
    meshinfo_dict, is_axi, ndim, _ = extract_hdf_meshinfo(meshfile)

    pdf_file = meshinfo_dict['Mesh name'].split('.')[0] + '.mesh.pdf'
    pdf = canvas.Canvas(pdf_file, pagesize=pagesize, verbosity=0)
    width, height = pagesize

    # Title section
    coords = [MARGIN, height - MARGIN]
    mesh_name = meshinfo_dict['Mesh name'].split('.')[0]
    _make_title(pdf, mesh_name, coords, *pagesize)

    pdf.line(*coords, width - MARGIN, coords[1])

    # Meshinfo section
    _make_meshinfo(pdf, meshinfo_dict, coords)

    pdf.line(*coords, width - MARGIN, coords[1])

    # MeshView section
    size_max = [width - 2*MARGIN, coords[1] - 6*cm]
    _make_meshview(pdf, meshfile, is_axi, ndim, coords, size_max)

    pdf.line(*coords, width - MARGIN, coords[1])

    # Histogram section
    size_max = [width - 2*MARGIN, coords[1] - MARGIN]
    _make_hist(pdf, meshfile, coords, size_max)

    # Main rectangle
    pdf.rect(MARGIN, MARGIN, width - 2*MARGIN, height - 2*MARGIN)

    print(f'{pdf_file} file saved.')
    pdf.save()

def generate_mesh_idcard(meshfile, pagesize):
    """
    *Main function distributing to landscape or portrait in function of the size*

    :param meshfile: Path to hdf5 mesh file
    :type meshfile: str
    :param pagesize: Tuple of page size (width, height) in pt (1 pt = 1/72 inch)
    """
    if pagesize[0] > pagesize[1]:
        generate_landscape(meshfile, pagesize)
    else:
        generate_portrait(meshfile, pagesize)



"""This module contains hip functionnalities
    intending to perform mesh operations
"""
from pyhip.hipster import pyhip_cmd

__all__ = ["generate_mesh_2d_3d",
           "adapt_with_factor",
           "adapt_with_var",
           "interpolate",
           "duplicate",
           "transform_translate",
           "transform_scale",
           "transform_rotate",
           "transform_reflect",
           "set_bctype",
           "set_bctext",
           "set_checklevel",
           "list_periodic",
           "list_surface",
           "hip_exit"]

def generate_mesh_2d_3d(lower_corner, upper_corner, resolution, convert2tri=False,
                        extru_axis=None, extru_range=None, extru_res=None, **kwargs):
    """
    Generate an unstructured mesh in 2D or 3D
    given bounding box.
    Can be converted from quat to tri or from hexa to tetra

    :param lower_corner: List/tuple (x_min, y_min) or (x_min, y_min, z_min)
    :param upper_corner: List/tuple (x_min, y_min) or (x_min, y_min, z_min)
    """
    cmds = []
    l_c = isinstance(lower_corner, (list, tuple)) and len(lower_corner) == 2
    u_c = isinstance(upper_corner, (list, tuple)) and len(upper_corner) == 2
    res = isinstance(resolution, (list, tuple)) and len(resolution) == 2
    if not (l_c and u_c and res):
        raise ValueError("Corners and resolution must be either lists or tuples"
                         "of 2 elements.")

    cmds += _generate_2d_mesh(lower_corner,
                              upper_corner,
                              resolution,
                              **kwargs)

    if all(var is not None for var in (extru_axis, extru_range, extru_res)):
        if not (isinstance(lower_corner, (list, tuple)) and len(lower_corner) == 2):
            raise ValueError("extru_range must be either list or tuple of 2 elements.")

        cmds += _extrude_2d_mesh(extru_range,
                                 extru_res,
                                 extru_axis,
                                 **kwargs)
        if convert2tri:
            cmds += _convert_hex2tet(**kwargs)
    elif convert2tri:
        cmds += _convert_quad2tri(**kwargs)

    return cmds

def _generate_2d_mesh(lower_corner, upper_corner, resolution, **kwargs):
    """Generate an unstructured rectangular mesh
       of quadrilateral elements given box bounds

       Parameters:
       ==========
       lower_corner: A tuple/list of box lower corner
                     coordinates (x, y)
       upper_corner: A tuple/list of box upper corner
                     coordinates (x, y)
       resolution : a tuple/list of mesh resolution
                    (i.e number of points) (n_x, n_y)
    """
    if lower_corner[0] > upper_corner[0] or lower_corner[1] > upper_corner[1]:
        raise ValueError(f"Lower corner coords {lower_corner} must be "
                         f"lower than upper corner coords {upper_corner}")

    command = (f"{lower_corner[0]} {lower_corner[1]} {upper_corner[0]}"
               f" {upper_corner[1]} {resolution[0]} {resolution[1]}")
    command = f"generate {command}"

    return pyhip_cmd(command, **kwargs)

def _extrude_2d_mesh(extrude_coords, extrude_res, axis, **kwargs):
    """ Extrude 2-dimensional mesh to a 3D mesh.

        Parameters:
        ==========
        extrude_coords: a tuple/list of extrusion extremes
                        (ext_min, ext_max), ext_min and ext_max
                        can refer to positions (if extusion is
                        done around x, y or z axis) or to angles
                        in degrees (if extusion is choosen to be
                        axisymmetric)
        extrude_node_num: Number of elements slices
        axis: axis around which extrusion is performed, possible
              values :
               - x, y, z or axi
    """
    if axis not in ['x', 'y', 'z', 'axi']:
        raise ValueError("extrusion axis must be in 'x', 'y', 'z', 'axi'")
    if extrude_coords[0] > extrude_coords[1]:
        raise ValueError(f"Lower coords {extrude_coords[0]} must be "
                         f"lower than upper coords {extrude_coords[1]}")

    command = f"copy 3D {extrude_coords[0]} {extrude_coords[1]} {extrude_res} {axis}"

    return pyhip_cmd(command, **kwargs)

def _convert_quad2tri(**kwargs):
    """ Convert current quad mesh to triangle elements.

        Parameters:
        ==========
        none
    """
    command = 'copy q2t'

    return pyhip_cmd(command, **kwargs)

def _convert_hex2tet(**kwargs):
    """ Convert current hex mesh to tet elements.

        Parameters:
        ==========
        none
    """
    command = 'copy 2tets'

    return pyhip_cmd(command, **kwargs)

def adapt_with_factor(factor, hgrad=1.4, hmin=0.0, **kwargs):
    """ Adapt mesh using a coefficient.

        Parameters:
        ==========
        isofactor : a double parameters by which every edge length
                    will be multiplied
        hgrad : *optional*, maximum gradient between edge sizes during adaptation
        Hdist : *optional*, Hausdorff maximum distance
        hmin : *optional*, minimal edge size
        hmax : *optional*, maximal edge size
    """
    command = f"mmg3d -f {factor}"
    if hgrad != 1.4:
        command += f" -g {hgrad}"
    if hmin != 0.0:
        command += f" -l {hmin}"

    return pyhip_cmd(command, **kwargs)

def adapt_with_var(variable, hgrad=1.4, hmin=None, **kwargs):
    """ Adapt mesh using a variable.

        Parameters:
        ==========
        variable : a variable name (string) contained in the current memory
        hGrad : *optional*, maximum gradient between edge sizes during adaptation
        Hdist : *optional*, Hausdorff maximum distance
        hmin : *optional*, minimal edge size
        hmax : *optional*, maximal edge size
    """
    command = f"mmg3d -v {variable}"
    if hgrad != 1.4:
        command += f" -g {hgrad}"
    if hmin is not None:
        command += f" -l {hmin}"

    return pyhip_cmd(command, **kwargs)

def interpolate(grid_id=1, **kwargs):
    """ interpolates solution to current mesh in memory
        Parameters:
        ==========
        GridId : id of the grid info to use as source (optional, default 1)
    """
    command = f"interpolate grid {grid_id}"

    return pyhip_cmd(command, **kwargs)

def duplicate(ncopies, value, operation='rotate', axis='x', **kwargs):
    """ copies a mesh (and solution) using the operation type with the values parameters

        Parameters:
        ==========
        ncopies : int number of copies
        axis : 'x', 'y', or 'z'
        angle : double value of the angle
    """
    if operation == 'rotate':
        if axis not in ['x', 'y', 'z']:
            raise ValueError("Duplication by rotation's axis must be in 'x', 'y', 'z'")
        if not isinstance(value, float):
            raise ValueError("Duplication by rotation's value must be a float angle")
    elif operation == 'translate':
        axis = ''
        if isinstance(value, list) and len(value) == 3:
            value = f"{value[0]} {value[1]} {value[2]}"
        else:
            raise ValueError("Duplication by translation value must be a 3-elements list")
    else:
        raise ValueError("Duplication operation must be 'rotate' or 'translate'")

    if isinstance(ncopies, int):
        command = f"copy uns {ncopies} {operation} {axis} {value}"
    else:
        raise ValueError("The number of copies (ncopies) must be an integer")

    return pyhip_cmd(command, **kwargs)

def transform_translate(d_x, d_y, d_z, **kwargs):
    """ translates mesh along a direction
        Parameters:
        ==========
        direction : tuple of 3 direction components (dx, dy, dz)
    """
    command = f"transform translate {d_x} {d_y} {d_z}"

    return pyhip_cmd(command, **kwargs)

def transform_scale(s_x, s_y, s_z, **kwargs):
    """ apply a scaling by factor in each direction
        Parameters:
        ==========
        factor : tuple of 3 scaling factor (sx, sy, sz)
    """
    command = f"transform scale {s_x} {s_y} {s_z}"

    return pyhip_cmd(command, **kwargs)

def transform_rotate(axis, angle, **kwargs):
    """ rotate a mesh around x, y or z-axis by an angle in degree
        The right hand rule is applied.
        Parameters:
        ==========
        axis : str 'x', 'y' or 'z' of the axis
        angle : float value of the angle
    """
    if axis in ['x', 'y', 'z']:
        command = f"transform rotate {axis} {angle}"
    else:
        raise ValueError("rotation axis must be in 'x', 'y', 'z'")

    return pyhip_cmd(command, **kwargs)

def transform_reflect(axis, **kwargs):
    """ reflect a mesh around a plane perpendicular to the axis at the origin
        Parameters:
        ==========
        axis : str 'x', 'y' or 'z' of the axis
    """
    if axis in ['x', 'y', 'z']:
        command = f"transform reflect {axis}"
    else:
        raise ValueError("duplication axis must be in 'x', 'y', 'z'")

    return pyhip_cmd(command, **kwargs)

def set_bctype(patch, bctype, **kwargs):
    """ set boundary condition type on a specific patch
        Parameters:
        ==========
        patch : str patch name
        bctype : BC type of the patch
    """
    command = f"set bc-type {patch} {bctype}"

    return pyhip_cmd(command, **kwargs)

def set_bctext(patch, bctext, **kwargs):
    """ set boundary condition type on a specific patch
        Parameters:
        ==========
        patch : str patch name
        bctype : BC type of the patch
    """
    command = "set bc-text %s %s" % (patch, bctext)

    return pyhip_cmd(command, **kwargs)

def set_checklevel(level, **kwargs):
    """ Set check level
    """
    if not isinstance(level, int):
        raise TypeError("the level given must be an integer between 0 and 5")
    if not -1 < level < 6:
        raise ValueError("Check levels are only between 0 and 5 included")
    command = f"set checklevel {level}"

    return pyhip_cmd(command, **kwargs)

def list_periodic(**kwargs):
    """ List periodicities if exist
    """
    command = "list periodic"

    return pyhip_cmd(command, **kwargs)

def list_surface(**kwargs):
    """ List surfaces if exist
    """
    command = "list surface"

    return pyhip_cmd(command, **kwargs)

def hip_exit(**kwargs):
    """ exit hip
    """
    command = "exit"

    return pyhip_cmd(command, **kwargs)



"""Module containing hip writers functions"""
import os
from pyhip.hipster import pyhip_cmd


__all__ = ["write_fieldview",
           "write_avbp",
           "write_gmsh",
           "write_ensight",
           "write_cgns",
           "write_hdf5",
           "dump_wired"]

def _write_file(rootfile, output_format, writer_options=None, **kwargs):
    parent_dir = os.path.dirname(os.path.abspath(rootfile))
    basename, _ = os.path.splitext(os.path.basename(os.path.abspath(rootfile)))
    options = ['write', output_format]
    os.makedirs(parent_dir, exist_ok=True)
    if writer_options is not None:
        options.extend(writer_options)
    options.append(basename)

    commands = ["set path %s"%parent_dir]
    commands.append(" ".join(options))
    for command in commands:
        _ = pyhip_cmd(command, **kwargs)

    return commands

def write_fieldview(rootfile, **kwargs):
    """ Write mesh into fieldview format

        Parameters:
        ==========
        rootfile : output file base name path (without extension)
    """
    return _write_file(rootfile, "fieldview", **kwargs)

def write_avbp(rootfile, flavour=None, level=None, **kwargs):
    """ Write mesh and optionnaly solution to avbp format

        Parameters:
        ==========
        rootfile : output file base name path (without extension)
        flavour: should be one of the avbp flavours :
                 "avad", "avbp4.2", "avbp4.7", "avbp5.1", "avbp5.3eg",
                 "avbp5.3", "avh"
        level : level of coarsed mesh to be written, default is finest
                mesh
    """
    flavours = ["avad", "avbp4.2", "avbp4.7", "avbp5.1", "avbp5.3eg",
                "avbp5.3", "avh"]
    options = []
    if flavour is not None and flavour in flavours:
        options.append(flavour)
    if level is not None:
        try:
            options.append("%d" %int(level))
        except:
            raise ValueError("write_avbp : level should be an integer")

    return _write_file(rootfile, "avbp", options, **kwargs)

def write_gmsh(rootfile, **kwargs):
    """ Write mesh into gmsh format

        Parameters:
        ==========
        rootfile : output file base name path (without extension)
    """
    return _write_file(rootfile, "gmsh", **kwargs)

def write_ensight(rootfile,
                  writing_ascii=False,
                  write_node_ids=False,
                  segments=False,
                  extrude_2d=True,
                  **kwargs):
    """ Write mesh int ensight gold format

        Parameters:
        ==========
        rootfile : output file base name path (without extension)
        writing_ascii : if True, ascii format is adopted
        write_node_ids : if True, store node ids
        segments : if true, writes only 1D elements in geofile instead of 2D/3D ones
        this ignores the extrude_2d
        extrude_2d : if True, 2D mesh is extruded to 3D
                     in order to be readable by paraview
    """
    options = []
    if writing_ascii:
        options.append('-a')
    if write_node_ids:
        options.append('-n on')
    if segments:
        options.append('-s0')
    elif extrude_2d:
        options.append('-3')
    else:
        options.append('-2')

    return _write_file(rootfile, "ensight", options, **kwargs)

def dump_wired(rootfile, **kwargs):
    """ dump a wired geometry into rootfile.geo and rootfile.case
        Parameters:
        ==========
        rootfile : str root of the files to dump
    """
    command1 = "decimate"
    cmd1 = pyhip_cmd(command1, **kwargs)

    cmd23 = write_ensight(rootfile, writing_ascii=True, segments=True, **kwargs)

    return cmd1 + cmd23

def write_cgns(rootfile, **kwargs):
    """ Write mesh into fieldview format

        Parameters:
        ==========
        rootfile : output file base name path (without extension)
    """
    return _write_file(rootfile, "cgns", **kwargs)

def write_hdf5(rootfile,
               write_all=True,
               only_solution=False,
               separate_boundary_shell=False,
               add_metis_graph=False,
               write_faces_list=False,
               compression_level=None,
               **kwargs):
    """ Write mesh and optionnaly solution to file

        Parameters:
        ==========
        rootfile : output file base name path (without extension)
        write_all : if True all the variables that are stored,
                    otherwize, only the standard set is stored
        separate_boundary_shell : if True, the boundary shell (skin)
                                  is written as a separately numbered
                                  mesh entity for each boundary patch.
        compression_level: sets the zip compression level.
                            Note that due to overhead, compression with
                            the hdf native routines does not always result
                            in relevant file size reduction, or at times
                            any reduction at all.
        add_metis_graph: If True adds the METIS style elGraph.
        write_faces_list : if True, a complete list of faces between
                            elements is written to file.
                            Warning: Writing the complete list of faces
                            will enlarge the file size considerably.
        only_solution: if True, write a solution only but not the mesh
    """
    #pylint: disable=unused-argument, too-many-arguments
    usr_opts = locals().copy()
    opt_values = {"write_all":["-a", "-a0"],
                  "only_solution":["-s", ""],
                  "separate_boundary_shell":["-b", ""],
                  "add_metis_graph":['-e', ''],
                  "write_faces_list":['-f', '']}
    options = ['-7']
    for opt in opt_values:
        value = opt_values[opt][1]
        if usr_opts[opt]:
            value = opt_values[opt][0]
        if value:
            options.append(value)

    if usr_opts["compression_level"] is not None:
        options.append("-c %d" % compression_level)
    return _write_file(rootfile, "hdf5", options, **kwargs)



"""
Startup script to call calculator gui  (**IHM**).
"""
import os
import inspect
import yaml
from opentea.gui_forms.otinker import main_otinker


def main(schema_file=None):
    """
    *Call the otinker gui.*
    """
    if schema_file is None:
        schema_file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                   "hip.yaml")
    with open(schema_file, 'r') as fin:
        schema = yaml.load(fin, Loader=yaml.FullLoader)
    base_dir = inspect.getfile(inspect.currentframe())
    base_dir = os.path.dirname(os.path.abspath(base_dir))
    main_otinker(schema, calling_dir=base_dir, tab_3d=True)

if __name__ == "__main__":
    main()



""" This module contains scripts for full basic scenario with pyhip"""
from pyhip.setup import (runner_setup,
                         read_mesh,
                         transform,
                         set_none_bctype,
                         periodize,
                         write_files,
                         process_end,
                         adapt)

__all__ = ["adaptation_setup"]


def adaptation_setup(nob_in, fallback, checklevel=True):
    """generic hip adaptation setup scenario"""
    hip_cmds = [runner_setup(fallback, checklevel=checklevel)]

    if 'read' in nob_in:
        hip_cmds += read_mesh(nob_in.read, fallback=fallback)
    else:
        raise RuntimeError("a 'read' block must be specified")

    #mesh operations
    if 'transform' in nob_in:
        hip_cmds += transform(nob_in.transform, fallback=fallback)

    #mesh periodization
    if 'periodize' in nob_in:
        hip_cmds += set_none_bctype(nob_in.patch_list, fallback=fallback)
        hip_cmds += periodize(nob_in, nob_in.patch_list, fallback=fallback)

    #mesh adaptations
    if 'adapt' in nob_in:
        hip_cmds += adapt(nob_in.adapt, fallback=fallback)
    else:
        raise RuntimeError("an 'adapt' block must be specified.")
    #mesh writing
    if 'write' in nob_in:
        hip_cmds += write_files(nob_in.write, fallback=fallback)
    else:
        raise RuntimeError("a 'write' block must be specified")

    hip_cmds += process_end(fallback)

    return hip_cmds



from tiny_3d_engine import load_file_as_scene
from nob.core import Tree
from opentea.process_utils import update_3d_callback
from opentea.noob.noob import unique_dict_key

def update_3d_scene(nob_in, scene):
    """Update the list of dimensions."""

    nob_tree = Tree(nob_in)
    if nob_in.interpolation.find('interpolation'):
        geo_path = nob_tree.interpolate.geofile[:]
        scene = load_file_as_scene(geo_path, prefix='interp', scene=scene, color="#0000ff")
        scene.add_axes()

    return scene


if __name__ == "__main__":
    print("3D CBck:")
    update_3d_callback(update_3d_scene)



"""
.. automodule:: hippy.readers
   :members:
"""
#pylint: disable=import-error,no-name-in-module
# from hip_wrapper import hip_cmd
from .hipster import *
from . import commands
from . import gui
from . import scenarii
from . import setup



#!/usr/bin/env python
"""
cli.py

Command line interface for tools in pyhip
"""

import click

FORMATS = ('hdf5', 'fluent', 'centaur', 'gmsh', 'cgns', 'ensight')

@click.group()
def main_cli():
    """---------------  PYHIP  --------------------

You are now using the Command line interface of PYHIP,
a Python3 helper to interact with the mesh management tool HIP
Pyhip was created at CERFACS (https://cerfacs.fr).

This is a python package currently installed in your python environement.
"""
    pass

@click.command()
def gui():
    """ Launch PyHIP IHM """

    import pyhip.gui.startup as gui_startup
    gui_startup.main()
main_cli.add_command(gui)

@click.command()
@click.argument('meshfile', type=click.Path(exists=True), nargs=1)
@click.option(
    '-f', '--format',
    type=click.Choice(['portrait', 'landscape'], case_sensitive=False),
    default='landscape')
def meshid(meshfile, format):
    """ Generate mesh ID Card """

    from pyhip.commands import generate_mesh_idcard
    format_dict = {
        'portrait': (595.2755905511812, 841.8897637795277),
        'landscape': (960., 480.)
    }
    pagesize = format_dict[format.lower()]
    generate_mesh_idcard(meshfile, pagesize)

main_cli.add_command(meshid)

@click.command()
@click.argument('meshfile', type=click.Path(exists=True), nargs=1)
def info(meshfile):
    """ Display mesh informations """

    from pyhip.commands.mesh_idcard import extract_hdf_meshinfo, arrange_meshinfo
    meshinfo_dict, _, _, patch_labels = extract_hdf_meshinfo(meshfile)
    meshinfo_dict['Patch labels'] = {str(i+1): patch for i, patch in enumerate(patch_labels)}
    meshinfo_as_lines = arrange_meshinfo(meshinfo_dict)
    
    print('\n'.join(meshinfo_as_lines))
main_cli.add_command(info)

@click.command()
@click.argument('mesh_source', type=click.Path(exists=True), nargs=1)
@click.argument('solution_source', type=click.Path(exists=True), nargs=1)
@click.argument('mesh_target', type=click.Path(exists=True), nargs=1)
@click.option('--checklevel', type=click.IntRange(0, 5, clamp=True), default=0)
@click.option('--fallback/--nofallback', default=True)
def interpolate(mesh_source, solution_source, mesh_target, checklevel, fallback):
    """ Interpolate a source solution from 
        a source mesh to a target_mesh """
    from pathlib import Path
    from pyhip.commands.readers import read_mesh_files
    from pyhip.commands.operations import interpolate, set_checklevel
    from pyhip.commands.writers import write_hdf5
    from pyhip.setup import process_end

    set_checklevel(checklevel, fallback=fallback)
    read_mesh_files([mesh_source, solution_source], 'hdf5', fallback=fallback)
    read_mesh_files([mesh_target], 'hdf5', fallback=fallback)
    interpolate(grid_id=1, fallback=fallback)
    write_hdf5(
       Path(mesh_target).name.split('.')[0],
       only_solution=True,
       fallback=fallback,
    )

    process_end(fallback=fallback)
main_cli.add_command(interpolate)

@click.command(context_settings=dict(ignore_unknown_options=True))
@click.argument('lower_corner', type=float, nargs=2)
@click.argument('upper_corner', type=float, nargs=2)
@click.argument('resolution', type=int, nargs=2)
@click.option('--tri', is_flag=True)
@click.option('--fallback/--nofallback', default=True)
def generate2d(lower_corner, upper_corner, resolution, tri, fallback):
    """ Generate 2d mesh """

    from pyhip.commands import (
        generate_mesh_2d_3d,
        write_hdf5,
    )
    from pyhip.setup import process_end

    generate_mesh_2d_3d(
        lower_corner,
        upper_corner,
        resolution,
        convert2tri=tri,
        fallback=fallback,
    )
    write_hdf5('mesh_2d', fallback=fallback)

    process_end(fallback=fallback)
main_cli.add_command(generate2d)

@click.command(context_settings=dict(ignore_unknown_options=True))
@click.argument('lower_corner', type=float, nargs=2)
@click.argument('upper_corner', type=float, nargs=2)
@click.argument('resolution', type=int, nargs=2)
@click.argument('extru_axis', type=click.Choice(['x', 'y', 'z', 'axi']), nargs=1)
@click.argument('extru_range', type=float, nargs=2)
@click.argument('extru_res', type=int, nargs=1)
@click.option('--tetra', is_flag=True)
@click.option('--fallback/--nofallback', default=True)
def generate3d(lower_corner, upper_corner, resolution, extru_axis,
               extru_range, extru_res, tetra, fallback):
    """ Generate 3d mesh """

    from pyhip.commands import (
        generate_mesh_2d_3d,
        write_hdf5,
    )
    from pyhip.setup import process_end

    generate_mesh_2d_3d(
        lower_corner,
        upper_corner,
        resolution,
        extru_axis=extru_axis,
        extru_range=extru_range,
        extru_res=extru_res,
        convert2tri=tetra,
        fallback=fallback,
    )
    write_hdf5('mesh_3d', fallback=fallback)

    process_end(fallback=fallback)
main_cli.add_command(generate3d)

@click.command()
@click.argument('meshfile', type=click.Path(exists=True), nargs=1)
@click.argument('factor', type=float, nargs=3)
@click.option('--checklevel', type=click.IntRange(0, 5, clamp=True), default=0)
@click.option('--fallback/--nofallback', default=True)
def scale(meshfile, factor, checklevel, fallback):
    """ Scale geometry by scaling coefficients vector """

    from pathlib import Path
    from pyhip.commands import (
        set_checklevel,
        read_mesh_files,
        transform_scale,
        write_hdf5,
    )
    from pyhip.setup import process_end

    set_checklevel(checklevel, fallback=fallback)
    read_mesh_files([meshfile], 'hdf5', fallback=fallback)
    transform_scale(*factor, fallback=fallback)
    write_hdf5(Path(meshfile).name.split('.')[0] + '_scaled', fallback=fallback)

    process_end(fallback=fallback)
main_cli.add_command(scale)

@click.command(context_settings=dict(ignore_unknown_options=True))
@click.argument('meshfile', type=click.Path(exists=True), nargs=1)
@click.argument('axis', type=click.Choice(('x', 'y', 'z'), case_sensitive=False), nargs=1)
@click.argument('angle', type=float, nargs=1)
@click.option('--checklevel', type=click.IntRange(0, 5, clamp=True), default=0)
@click.option('--fallback/--nofallback', default=True)
def rotate(meshfile, axis, angle, checklevel, fallback):
    """ Rotate geometry of angle [deg] around axis """

    from pathlib import Path
    from pyhip.commands import (
        set_checklevel,
        read_mesh_files,
        transform_rotate,
        write_hdf5,
    )
    from pyhip.setup import process_end

    set_checklevel(checklevel, fallback=fallback)
    read_mesh_files([meshfile], 'hdf5', fallback=fallback)
    transform_rotate(axis, angle, fallback=fallback)
    write_hdf5(Path(meshfile).name.split('.')[0] + '_rotated', fallback=fallback)

    process_end(fallback=fallback)
main_cli.add_command(rotate)

@click.command(context_settings=dict(ignore_unknown_options=True))
@click.argument('meshfile', type=click.Path(exists=True), nargs=1)
@click.argument('direction', type=float, nargs=3)
@click.option('--checklevel', type=click.IntRange(0, 5, clamp=True), default=0)
@click.option('--fallback/--nofallback', default=True)
def translate(meshfile, direction, checklevel, fallback):
    """ Translate geometry by direction vector """

    from pathlib import Path
    from pyhip.commands import (
        set_checklevel,
        read_mesh_files,
        transform_translate,
        write_hdf5,
    )
    from pyhip.setup import process_end

    set_checklevel(checklevel, fallback=fallback)
    read_mesh_files([meshfile], 'hdf5', fallback=fallback)
    transform_translate(*direction, fallback=fallback)
    write_hdf5(Path(meshfile).name.split('.')[0] + '_translated', fallback=fallback)

    process_end(fallback=fallback)
main_cli.add_command(translate)

@click.command()
@click.argument('meshfile', type=click.Path(exists=True), nargs=1)
@click.argument('format_in', type=click.Choice(FORMATS, case_sensitive=False), nargs=1)
@click.argument('format_out', type=click.Choice(FORMATS, case_sensitive=False), nargs=1)
@click.option('--checklevel', type=click.IntRange(0, 5, clamp=True), default=0)
@click.option('--fallback/--nofallback', default=True)
def convert(meshfile, format_in, format_out, checklevel, fallback):
    """ Convert a meshfile into specific format mesh """

    from pathlib import Path
    from pyhip.commands import (
        set_checklevel,
        read_mesh_files,
        write_hdf5,
    )
    from pyhip.setup import process_end

    if format_out.lower() != 'hdf5':
        raise NotImplementedError(f"Format '{format_out} not implemented yet.")

    set_checklevel(checklevel, fallback=fallback)
    read_mesh_files([meshfile], format_in.lower(), fallback=fallback)
    if format_out.lower() == 'hdf5':
        write_hdf5(Path(meshfile).name.split('.')[0], fallback=fallback)

    process_end(fallback=fallback)
main_cli.add_command(convert)

@click.command()
def interactive():
    """ Launch bash hip.exe """
    import os
    import platform
    import pkg_resources

    hip_cmd = pkg_resources.resource_filename("pyhip", "hip_%s.exe" % platform.system())
    os.system(hip_cmd)
main_cli.add_command(interactive)

@click.command()
@click.argument('script', type=click.Path(exists=True), nargs=1)
def script(script):
    """ Excecute hip script in batch mode """

    import os
    import platform
    import pkg_resources

    with open(script, 'a') as fscript:
        fscript.write('exit\n')

    hip_cmd = pkg_resources.resource_filename("pyhip", "hip_%s.exe" % platform.system())
    os.system(f'{hip_cmd} {script}')
main_cli.add_command(script)

@click.command()
@click.argument('meshfile', type=click.Path(exists=True), nargs=1)
@click.argument('factor', type=float, nargs=1)
@click.option('--checklevel', type=click.IntRange(0, 5, clamp=True), default=0)
@click.option('--fallback/--nofallback', default=True)
def coarsen(meshfile, factor, checklevel, fallback):
    """ Coarsen a mesh with isoFactor > 1 """

    from pathlib import Path
    from pathlib import Path
    from pyhip.commands import (
        set_checklevel,
        read_mesh_files,
        adapt_with_factor,
        write_hdf5,
    )
    from pyhip.setup import process_end

    if factor <= 1.:
        msg = "Coarsening factor must be greater than 1."
        raise IOError(msg)

    set_checklevel(checklevel, fallback=fallback)
    read_mesh_files([meshfile], 'hdf5', fallback=fallback)
    adapt_with_factor(factor, fallback=fallback)
    write_hdf5(Path(meshfile).name.split('.')[0] + '_coarsened', fallback=fallback)

    process_end(fallback=fallback)
main_cli.add_command(coarsen)

@click.command()
@click.argument('meshfile', type=click.Path(exists=True), nargs=1)
@click.argument('factor', type=float, nargs=1)
@click.option('--checklevel', type=click.IntRange(0, 5, clamp=True), default=0)
@click.option('--fallback/--nofallback', default=True)
def refine(meshfile, factor, checklevel, fallback):
    """ Refine a mesh with isoFactor < 1 """

    from pathlib import Path
    from pyhip.commands import (
        set_checklevel,
        read_mesh_files,
        adapt_with_factor,
        write_hdf5,
    )
    from pyhip.setup import process_end

    if factor >= 1.:
        msg = "Refining factor must be lower than 1."
        raise IOError(msg)
    
    set_checklevel(checklevel, fallback=fallback)
    read_mesh_files([meshfile], 'hdf5', fallback=fallback)
    adapt_with_factor(factor, fallback=fallback)
    write_hdf5(Path(meshfile).name.split('.')[0] + '_refined', fallback=fallback)

    process_end(fallback=fallback) 
main_cli.add_command(refine)

@click.command()
@click.argument('meshfile', type=click.Path(exists=True), nargs=1)
@click.option('--checklevel', type=click.IntRange(0, 5, clamp=True), default=0)
@click.option('--fallback/--nofallback', default=True)
def improve(meshfile, checklevel, fallback):
    """ Improve mesh using HIP algorithm """

    from pathlib import Path
    from pyhip.commands import (
        set_checklevel,
        read_mesh_files,
        adapt_with_factor,
        write_hdf5,
    )
    from pyhip.setup import process_end
    
    set_checklevel(checklevel, fallback=fallback)
    read_mesh_files([meshfile], 'hdf5', fallback=fallback)
    adapt_with_factor(1., fallback=fallback)
    write_hdf5(Path(meshfile).name.split('.')[0] + '_improved', fallback=fallback)

    process_end(fallback=fallback)
main_cli.add_command(improve)

@click.command()
@click.argument('meshfile', type=click.Path(exists=True), nargs=1)
@click.option('--checklevel', type=click.IntRange(0, 5, clamp=True), default=0)
@click.option('--fallback/--nofallback', default=True)
def geo(meshfile, checklevel, fallback):
    """ Generate .geo and .case file """

    from pathlib import Path
    from pyhip.commands import(
        set_checklevel,
        read_mesh_files,
        dump_wired,
    )
    from pyhip.setup import process_end

    set_checklevel(checklevel, fallback=fallback)
    read_mesh_files([meshfile], 'hdf5', fallback=fallback) 
    root = Path(meshfile).parent / Path(meshfile).name.split('.')[0] 
    dump_wired(root, fallback=fallback)
    
    process_end(fallback=fallback)
main_cli.add_command(geo)

@click.command()
@click.argument('meshfiles', type=click.Path(exists=True), nargs=-1)
@click.option(
    "--shading",
    type=click.Choice(['none', 'linear', 'radial', 'flat' ], case_sensitive=False),
    default="flat",
    )
@click.option('--checklevel', type=click.IntRange(0, 5, clamp=True), default=0)
@click.option('--fallback/--nofallback', default=True)
@click.pass_context
def view(ctx, meshfiles, shading, checklevel, fallback):
    """ Quick mesh graphical view """

    import os
    from pathlib import Path
    from tiny_3d_engine import Engine3D, load_file_as_scene

    # White, Red, Blue, Green, Yellow
    colors = ('#ffffff', '#FF2D00', '#0013FF', '#00BF0C', '#FFF000')
    if len(meshfiles) > 5:
        msg = "View can not accept more than 5 meshes in a row"
        raise IOError(msg)

    scene = None
    for meshfile, color in zip(meshfiles, colors):
        prefix = Path(meshfile).name.split('.')[0] 
        geofile = Path(meshfile).parent / f'{prefix}.geo'
        if not geofile.exists():
            ctx.invoke(
                geo,
                meshfile=meshfile,
                checklevel=checklevel,
                fallback=fallback,
            )
        scene = load_file_as_scene(
            geofile,
            prefix=prefix,
            scene=scene,
            color=color,
        )
        scene.add_axes()

    engine = Engine3D(scene, shading=shading)
    engine.clear()
    engine.mainloop()
main_cli.add_command(view)





"""Module for the first tab."""
from nob import Nob
import os
from glob import glob
import h5py
from opentea.process_utils import process_tab
from opentea.noob.noob import unique_dict_key
from pyhip.commands.readers import read_mesh_files
from pyhip.commands.writers import (
    write_hdf5,
    dump_wired,
    )
from pyhip.commands.operations import (
    list_periodic,
    set_bctype,
    hip_exit,
    )
from pyhip.gui.parser_utils import (
    get_patch_list_AVBP,
    parse_meshinfo,
    parse_versioninfo,
    )

def process_interpolate(nob_in):
    """
    *Update mesh infos and show them in the IHM like this :*

    :param nob_in: A dict( ) with IHM's parameters

    :returns: A dict( ) with some update infos:

                - *patches_list*
    """
    nob_out = Nob(nob_in.copy())
    fallback = True # nob_out.fallback[:]
    msgs = []

    if unique_dict_key(nob_out.interpolation[:]) == 'No interpolation':
        return nob_out[:]

    # Reading of the mesh
    _ = read_mesh_files([nob_out.src_mesh[:]], "hdf5", fallback=fallback)
    
    # Récupération du filaire -> moteur 3D
    geo = nob_out.src_mesh[:].split('/')[-1].split('.')[0]
    _, _ = dump_wired(geo, fallback=fallback)
    nob_out.interpolate.geofile.set("./" + geo + ".geo")

    # Exit/Execute hip
    log, _ = hip_exit(fallback=fallback)

    # Setting of patch list
    patch_list = get_patch_list_AVBP(nob_out.src_mesh[:])
    nob_out.interpolate.patch_list_src.set(patch_list)

    if fallback:
        version_info = parse_versioninfo(log)
        mesh_info = parse_meshinfo(log)
        nob_out.interpolate.meshinfo.set(version_info + mesh_info)
    else:
        nob_out.interpolate.meshinfo.set("Mesh info parsing cython not implemented yet.")

    return nob_out[:]

if __name__ == "__main__":
    process_tab(process_interpolate)




"""hipster fallback for pyhip"""
import os
import subprocess
import platform
import pkg_resources
#pylint: disable=import-error,no-name-in-module
# from hip_wrapper import hip_cmd


HIP_CMD = pkg_resources.resource_filename("pyhip", "hip_%s.exe" % platform.system())

__all__ = ["pyhip_cmd"]

def pyhip_cmd(command, fallback=True):
    """Fallback function for pyhip when hip process fails"""
    # if not fallback:
    #     status, msg = hip_cmd(command)
    #     if status != 0:
    #         raise RuntimeError(msg)

    #     if msg is not None:
    #         print(msg)
    #         if command == "list periodic":
    #             return msg, [command]
    #         return [command]
    #     return [command]

    #fallback
    hip_file = os.path.abspath(f"{os.getpid()}_hip.in")

    with open(hip_file, 'a') as fout:
        fout.write(command + '\n')

    if command == "exit":
        status, log = run_command(hip_file)

        if status > 0:
            log_err = f"HIP process failed with exit code {status}.\n\n"
            log_err += "HIP commands :\n  "
            with open(hip_file, 'r') as fin:
                log_err += "  ".join(fin.readlines())
            os.remove(hip_file)
            log_err += "\nHIP error :\n  "
            with open('hip-fatal.log', 'r') as fin:
                log_err += "  ".join(fin.readlines())

            raise RuntimeError(log_err)

        os.remove(hip_file)
        return log, ['exit']
    return [command]

def run_command(hip_file):
    """ Run command """
    import shlex
    process = subprocess.Popen(
        shlex.split(f'{HIP_CMD} {hip_file}'),
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=os.path.dirname(hip_file),
    )
    log = []
    while process.poll() is None:
        stdout = process.stdout.readline()
        if stdout:
            log.append(stdout.decode('utf-8').rstrip())
            print(log[-1])
            
    return process.poll(), log



""" this module contains scripts for basic scenario usage"""
import warnings
from pyhip.commands.readers import read_mesh_files
from pyhip.commands.writers import (write_hdf5,
                                    write_avbp,
                                    write_ensight,
                                    write_cgns,
                                    write_gmsh,
                                    write_fieldview,
                                    dump_wired)
from pyhip.commands.operations import (transform_scale,
                                       transform_rotate,
                                       transform_translate,
                                       duplicate,
                                       generate_mesh_2d_3d,
                                       set_bctype,
                                       set_bctext,
                                       set_checklevel,
                                       hip_exit)

__all__ = ["runner_setup",
           "read_mesh",
           "generate_mesh",
           "transform",
           "set_none_bctype",
           "periodize",
           "write_files",
           "process_end"]


warnings.filterwarnings('always', category=UserWarning)

def runner_setup(fallback, checklevel=True):
    """runnner setup routine"""
    if fallback:
        warnings.warn_explicit("The fallback version of pyhip is being used.",
                               UserWarning,
                               'setup/core.py',
                               34)

    if checklevel:
        return set_checklevel(5, fallback=fallback)

    return set_checklevel(0, fallback=fallback)


def read_mesh(nob_in, **kwargs):
    """method to read different type of meshes in function of the nested object input"""
    return read_mesh_files(nob_in.files[:], nob_in.meshtype[:], **kwargs)

def transform(nob_in, **kwargs):
    """basic transformations"""
    cmds = []
    if nob_in.duplicate.nbr_sector_tgt[:] > 1:
        angle = 360. / nob_in.duplicate.nbr_sector_tot[:]
        cmds.extend(duplicate(nob_in.duplicate.nbr_sector_tgt[:] - 1, angle, **kwargs))

    if nob_in.scale.scaling_factor[:] != [1., 1., 1.]:
        cmds.extend(transform_scale(nob_in.scale.scaling_factor[:][0],
                                    nob_in.scale.scaling_factor[:][1],
                                    nob_in.scale.scaling_factor[:][2],
                                    **kwargs))

    for axis, angle in nob_in.rotate.angles[:].items():
        if angle != 0.:
            cmds.extend(transform_rotate(axis, angle, **kwargs))

    if nob_in.translate.direction[:] != [0., 0., 0.]:
        cmds.extend(transform_translate(nob_in.translate.direction[:][0],
                                        nob_in.translate.direction[:][1],
                                        nob_in.translate.direction[:][2],
                                        **kwargs))

    return cmds

def set_none_bctype(patch_list, **kwargs):
    """method to remove all boundary condition and set type to 'none'"""
    cmds = []
    for patch in patch_list:
        cmds += set_bctype(patch, 'n', **kwargs)

    return cmds

def set_bcname(new_patch_list, patch_list=None, patch_nbr=None, **kwargs):
    """method to set boundary condition names from number or name"""
    cmds = []
    if patch_list is None:
        if patch_nbr is None:
            raise RuntimeError("Either bc name or number must be provided"
                               "to set new patch names")
        for nbr, new_patch in zip(patch_nbr, new_patch_list):
            cmds += set_bctext(nbr, new_patch, **kwargs)
    else:
        for patch, new_patch in zip(patch_list, new_patch_list):
            cmds += set_bctext(patch, new_patch, **kwargs)

    return cmds

def generate_mesh(nob_in, **kwargs):
    """generate 2D or 3D mesh"""

    val = lambda n, v: n[n.find(v)[0]][:] if n.find(v) else None

    cmds = generate_mesh_2d_3d(
        nob_in.lower_corner[:],
        nob_in.upper_corner[:],
        nob_in.resolution[:],
        nob_in.convert2tri[:],
        extru_axis=val(nob_in, 'extru_axis'),
        extru_range=val(nob_in, 'extru_range'),
        extru_res=val(nob_in, 'extru_resolution'),
        **kwargs)

    return cmds

def periodize(nob_in, patch_list, **kwargs):
    """periodization of patch pairs for interpolated mesh with solution"""
    cmds = []
    if len(nob_in.periodize) > 0:
        for idx, pair in enumerate(nob_in.periodize[:]):
            if 'leader' in pair and 'follower' in pair:
                if pair['leader'] not in patch_list:
                    msg = f"Leader patch '{pair['leader']}' not in given patch list :"
                    msg += "\n - "
                    msg += "\n - ".join(patch_list)
                    raise ValueError(msg)
                if pair['follower'] not in patch_list:
                    msg = f"Follower patch '{pair['follower']}' not in given patch list :"
                    msg += "\n - "
                    msg += "\n - ".join(patch_list)
                    raise ValueError(msg)
                cmds += set_bctype(pair['leader'], f'u{idx:02}', **kwargs)
                cmds += set_bctype(pair['follower'], f'l{idx:02}', **kwargs)
            else:
                raise KeyError("A 'leader' and a 'follower' patch must be specified.")
    else:
        warnings.warn_explicit(('All patches will see their kind set to none,'
                                'all periodicities removed.'),
                               UserWarning,
                               'setup/core.py',
                               105)

    return cmds

def write_files(nob_in, **kwargs):
    """method to read different type of meshes in function of the nested object input"""
    # cgns, ensight, fluent, gmsh, hdf5
    msg = ("The only writing file types implemented are 'hdf5', 'avbp', 'ensight', ",
           "'cgns', 'gmsh' and 'fieldview'")
    if nob_in.casetype[:] == 'hdf5':
        cmds = write_hdf5(nob_in.casename[:], **nob_in.options[:], **kwargs)
    elif nob_in.casetype[:] == 'avbp':
        cmds = write_avbp(nob_in.casename[:], **nob_in.options[:], **kwargs)
    elif nob_in.casetype[:] == 'ensight':
        cmds = write_ensight(nob_in.casename[:], **nob_in.options[:], **kwargs)
    elif nob_in.casetype[:] == 'cgns':
        cmds = write_cgns(nob_in.casename[:], **kwargs)
    elif nob_in.casetype[:] == 'gmsh':
        cmds = write_gmsh(nob_in.casename[:], **kwargs)
    elif nob_in.casetype[:] == 'fieldview':
        cmds = write_fieldview(nob_in.casename[:], **kwargs)
    else:
        raise NotImplementedError(msg)

    cmds.extend(dump_wired(nob_in.casename[:], **kwargs))

    return cmds

def process_end(fallback):
    """simple function for exiting pyhip in scenarii"""
    hip_exit(fallback=fallback)
    print("\nPyhip successfuly exited.\n")

    return ['exit']



""" This module contains scripts for full basic scenario with pyhip"""
from pyhip.setup import (runner_setup,
                         read_mesh,
                         generate_mesh,
                         transform,
                         set_none_bctype,
                         periodize,
                         write_files,
                         process_end,
                         adapt,
                         interpolation,
                         remove_duplicated_bctext)
from pyhip.commands import generate_mesh_idcard
#from reportlab.lib.pagesizes import A4
LANDSCAPE = (960., 480.)

__all__ = ["complete_setup"]

def complete_setup(nob_in, fallback=True, checklevel=True):
    """generic hip complete setup scenario, the nested object in input must be of the Nob class"""
    hip_cmds = runner_setup(fallback, checklevel=checklevel)

    #meshes reading
    if 'interpolate' in nob_in:
        hip_cmds += read_mesh(nob_in.interpolate, fallback=fallback)
        hip_cmds += remove_duplicated_bctext(nob_in, fallback=fallback)

    if 'read' in nob_in:
        if 'generate' in nob_in:
            raise RuntimeError("Can't 'read' and 'generate'"
                               "blocks be both specified")
        hip_cmds += read_mesh(nob_in.read, fallback=fallback)
        patch_list = nob_in.read.patch_list[:]
    elif 'generate' in nob_in:
        hip_cmds += generate_mesh(nob_in.generate, fallback=fallback)
        patch_list = nob_in.generate.patch_list[:]
    else:
        raise RuntimeError("A 'read' or 'generate' block must be specified")

    #mesh operations
    if 'transform' in nob_in:
        hip_cmds += transform(nob_in.transform, fallback=fallback)

    #mesh periodization
    if 'periodize' in nob_in:
        if 'interpolate' in nob_in:
            hip_cmds += set_none_bctype(nob_in.interpolate.patch_list[:], fallback=fallback)
        hip_cmds += set_none_bctype(patch_list, fallback=fallback)
        hip_cmds += periodize(nob_in, patch_list, fallback=fallback)

    #solution interpolation
    if 'interpolate' in nob_in:
        hip_cmds += interpolation(fallback=fallback)

    #mesh adaptations
    if 'adapt' in nob_in:
        hip_cmds += adapt(nob_in.adapt, fallback=fallback)

    #mesh writing
    if 'write' in nob_in:
        hip_cmds += write_files(nob_in.write, fallback=fallback)
    else:
        raise RuntimeError("a 'write' block must be specified")

    hip_cmds += process_end(fallback)

    if nob_in.write.casetype[:] == 'hdf5':
        generate_mesh_idcard(
            nob_in.write.casename[:] + '.mesh.h5',
            LANDSCAPE)

    return hip_cmds



"""Module for the first tab."""
from nob import Nob
import os
from glob import glob
import h5py
from opentea.process_utils import process_tab
from opentea.noob.noob import unique_dict_key
from pyhip.gui.parser_utils import (
    get_patch_list_AVBP,
    parse_meshinfo,
    parse_versioninfo,
    parse_boundaries,
    parse_periodicities,
    )
from pyhip.commands.readers import read_mesh_files
from pyhip.commands.writers import (
    write_hdf5,
    dump_wired,
    )
from pyhip.commands.operations import (
    list_periodic,
    list_surface,
    generate_mesh_2d_3d,
    set_bctype,
    set_checklevel,
    hip_exit,
    )

def process_mesh(nob_in):
    """
    *Update mesh infos and show them in the IHM like this :*

    :param nob_in: A dict( ) with IHM's parameters

    :returns: A dict( ) with some update infos:

                - *patches_list*
    """
    nob_out = Nob(nob_in.copy())

    fallback = True # nob_out.fallback[:]
    if nob_out.checklvl[:]:
        set_checklevel(5, fallback=fallback)
    else:
        set_checklevel(0, fallback=fallback)

    if unique_dict_key(nob_out.input_mode[:]) == 'generate':
        if nob_out.generate.structure[:] == 'quad':
            convert2tri = False
        else:
            convert2tri = True

        if unique_dict_key(nob_out.dimension[:]) == '2d':
            extru_axis = None
            extru_range = None
            extru_res = None
        else:
            extru_range = nob_out.w_bounds[:]
            extru_res = nob_out.resolution[:][-1] - 1
            if unique_dict_key(nob_out.dimension[:]) == '3d_xyz':
                extru_axis = 'z'
            else:
                extru_axis = 'axi'

        generate_mesh_2d_3d(
            (nob_out.u_bounds[:][0], nob_out.v_bounds[:][0]),
            (nob_out.u_bounds[:][1], nob_out.v_bounds[:][1]),
            nob_out.resolution[:][:2],
            convert2tri=convert2tri,
            extru_axis=extru_axis,
            extru_range=extru_range,
            extru_res=extru_res,
            fallback=fallback)

    elif unique_dict_key(nob_out.input_mode[:]) == 'read':
        input_case = unique_dict_key(nob_out.input_case[:])
        case_fmt = input_case.split('_')[0]

        # Reading of the mesh
        mesh_file = nob_out.read.meshfile[:]
        _ = read_mesh_files([mesh_file], case_fmt, fallback=fallback)

        # Listing of periodicities
        _ = list_periodic(fallback=fallback)

    _ = list_surface(fallback=fallback)

    root = 'tmp'
    # Writing of the mesh in hdf5 format
    _ = write_hdf5(root, fallback=fallback)

    # Récupération du filaire -> moteur 3D
    geo = 'geo_tmp'
    _ = dump_wired(geo, fallback=fallback)
    nob_out.inputs.geofile.set("./" + geo + ".geo")

    # Exit/Execute hip
    log, _ = hip_exit(fallback=fallback)

    # Setting of patch list
    if fallback:
        patch_dict = parse_boundaries(log)
        patch_list = list(patch_dict.keys())
    else:
        patch_list = get_patch_list_AVBP(root + ".mesh.h5")
    nob_out.input_mode.patch_list.set(patch_list)

    # Setting of version and mesh info
    if fallback:
        version_info = parse_versioninfo(log)
        mesh_info = parse_meshinfo(log)
        nob_out.inputs.meshinfo.set(version_info + mesh_info)
    else:
        nob_out.inputs.meshinfo.set("Mesh info parsing cython not implemented yet.")

    # Setting of existed periodicity
    if fallback:
        perio_list = parse_periodicities(log)
        nob_out.periodicity.bnd_patch.set(perio_list)
    else:
        print('Periodicities parsing cython not implemented yet')

    for root_file in glob(root + '*'):
        os.remove(root_file)

    return nob_out[:]

if __name__ == "__main__":
    process_tab(process_mesh)




"""this module is for optional scenarii routines"""
from pyhip.commands.operations import (adapt_with_factor,
                                       set_bctext,
                                       interpolate)

__all__ = ["remove_duplicated_bctext",
           "interpolation",
           "adapt"]


def remove_duplicated_bctext(nob_in, **kwargs):
    """method to remove all boundary condition and set type to 'none'"""
    cmds = []
    for patch in nob_in.read.patch_list[:]:
        if patch in nob_in.interpolate.patch_list[:]:
            cmds += set_bctext(patch, patch + '_src', **kwargs)
    return cmds

def interpolation(**kwargs):
    """interpolate solution on current mesh"""
    return interpolate(grid_id=1, **kwargs)

def adapt(nob_in, **kwargs):
    """method to refine or coarse a mesh"""
    if 'hGrad' not in nob_in.factor:
        print("Default value of hGrad is used: 1.4")
        nob_in.factor['hGrad'] = 1.4
    if 'hmin' not in nob_in.factor:
        print("No minimal distance is targeted")
        nob_in.factor['hmin'] = 0.0

    cmds = adapt_with_factor(nob_in.factor.coefficient[:],
                             hgrad=nob_in.factor.hGrad[:],
                             hmin=nob_in.factor.hmin[:],
                             **kwargs)
    return cmds



"""Setup file for pyhip building/installing"""
from glob import glob
import os
import sys
import platform
from distutils.sysconfig import parse_makefile
from setuptools import setup, find_packages#, Extension

NAME = "pyhip"
VERSION = "0.3.1"
DESCR = """A Python interface to Hip which is a package for manipulating
           unstructured computational grids and their associated datasets"""
URL = "http://www.cerfacs.fr/avbp7x/hip.php"

AUTHOR = "Jens-Dominik Mueller, CERFACS"
EMAIL = "j.mueller@qmul.ac.uk, gabriel.staffelbach@cerfacs.fr, erraiya@cerfacs.fr"

LICENSE = "CeCILL-B Free Software License Agreement (CECILL-B)"

SRC_DIR = "pyhip"
PACKAGES = [SRC_DIR]
README = open(os.path.join(os.path.dirname(__file__), 'README.md')).read()
# EXT = build_extension()

setup(name=NAME,
      version=VERSION,
      description=DESCR,
      long_description=README,
      long_description_content_type='text/markdown',
      author=AUTHOR,
      author_email=EMAIL,
      url=URL,
      license=LICENSE,
      packages=find_packages("src"),
      package_dir={"": "src"},
      py_modules=[os.path.splitext(os.path.basename(path))[0] for path in glob('src/*.py')],
      include_package_data=True,
      #data_files = [('bin', ['./src/pyhip/hip_Linux.exe','./src/pyhip/hip_Darwnin.exe'])],
      zip_safe=False,
      python_requires='>=3.6.0',
      classifiers=["Development Status :: 4 - Beta",
                   "Programming Language :: Python :: 3.6",
                   "Programming Language :: Python :: 3.7",
                   "Programming Language :: Python :: 3.8",
                   "Programming Language :: Python :: 3.9",
                   "Topic :: Scientific/Engineering :: Physics"
                   ],
      setup_requires=[#"Cython",
                      "sdist",
                      "wheel",
                      "numpy",
                      "nob",
                      'reportlab',
                      "PyYAML",
                      "matplotlib",
                      "scipy",
                      "arnica",
                      "opentea>=3.1.0"],
      install_requires=["numpy",
                        "nob",
                        "reportlab",
                        "PyYAML",
                        "matplotlib",
                        "scipy",
                        "arnica",
                        "opentea>=3.1.0"],
      entry_points={
        "console_scripts": [
            "ihm_hip = pyhip.gui.startup:main",
            "pyhip = pyhip.cli:main_cli",
            ]
      }
      )



""" script to run scenarii for pyhip"""
from pyhip.scenarii import complete_setup, interpolation_setup, adaptation_setup, basic_setup
import yaml
from nob import Nob


def basic_scenario():
    """basic scenario launcher """
    with open('basic.yml', 'r') as yml_file:
        nob_in = Nob(yaml.load(yml_file, Loader=yaml.FullLoader))

    _ = basic_setup(nob_in, nob_in.fallback[:])

def complete_scenario():
    """complete scenario launcher """
    with open('complete.yml', 'r') as yml_file:
        nob_in = Nob(yaml.load(yml_file, Loader=yaml.FullLoader))

    _ = complete_setup(nob_in, nob_in.fallback[:], nob_in.checklevel[:])

def interpolation_scenario():
    """interpolation scenario launcher """
    with open('interpolation.yml', 'r') as yml_file:
        nob_in = Nob(yaml.load(yml_file, Loader=yaml.FullLoader))

    _ = interpolation_setup(nob_in, nob_in.fallback[:])

def adapt_scenario():
    """adapt scenario launcher """
    with open('adapt.yml', 'r') as yml_file:
        nob_in = Nob(yaml.load(yml_file, Loader=yaml.FullLoader))

    _ = adaptation_setup(nob_in, nob_in.fallback[:])

if __name__ == "__main__":
    """uncomment the ones you wish to use, repolace the in correct order if needed"""
    #basic_scenario()
    complete_scenario()
    #interpolation_scenario()
    #adapt_scenario()




""" Module executed when the **"process finish"** tab button is pressed."""
from pyhip.scenarii.complete import complete_setup
from opentea.process_utils import process_tab
from opentea.noob.noob import unique_dict_key
from nob import Nob
import yaml

def process_finish(nob_in):
    """
    *Update the inlet dict( ) and make the different setups with the*  \
    :py:func:`pyavbp.scenarii.tavbp_setup.tavbp`  *function.*

    :param nob_in: a dict( ) with IHM's parameters

    :returns: A **dict( )** with some update infos:

    """
    nob_out = Nob(nob_in.copy())

    complete_setup_data(nob_out)

    return nob_out[:]

def complete_setup_data(nob):
    """method to setup a dictionary to be read by the basic scenario"""

    setup_nob = dict()
    setup_nob.update(_inputs(nob.input_mode))
    setup_nob.update(_periodize(nob.periodicity))
    setup_nob.update(_transform(nob.transform))
    setup_nob.update(_interpolate(nob.interpolation))
    setup_nob.update(_adapt(nob.adaptation))
    setup_nob.update(_write(nob.write))

    setup_nob['runner'] = {'fallback': True, #nob.fallback[:], 
                           'checklevel': nob.checklvl[:]}

    with open('ihm_setup.yml', 'w') as fout:
        yaml.dump(setup_nob, fout, default_flow_style=False)

    hip_commands = complete_setup(
        Nob(setup_nob),
        fallback=True, #nob.fallback[:],
        checklevel=nob.checklvl[:],
    )

    nob.hip_cmd.set("\n".join(hip_commands))

def _inputs(nob):
    """method to setup input parameters"""

    if 'read' in nob:
        out = _read(nob.read)
    elif 'generate' in nob:
        out = _generate(nob.generate)

    return out

def _read(nob):
    """method to setup read parameters"""

    input_case = unique_dict_key(nob.input_case[:])
    case_fmt = input_case.split('_')[0]
    if 'solution' in input_case:
        avefile = nob.avefile[:]
    else:
        avefile = None

    out = {
        'files': [nob.meshfile[:], avefile],
        'meshtype': case_fmt,
        'patch_list': nob.patch_list[:],
    }

    return {'read': out}

def _generate(nob):
    """method to setup generate parameters"""

    out = {}
    if nob.structure[:] == 'quad':
        convert2tri = False
    else:
        convert2tri = True

    out = {
        'lower_corner': (nob.u_bounds[:][0], nob.v_bounds[:][0]),
        'upper_corner': (nob.u_bounds[:][1], nob.v_bounds[:][1]),
        'resolution': nob.resolution[:][:2],
        'convert2tri': convert2tri,
        'patch_list': nob.patch_list[:],
    }
    if unique_dict_key(nob.dimension[:]) in ('3d_xyz', '3d_xrtheta'):
        out['extru_range'] = nob.w_bounds[:]
        out['extru_resolution'] = nob.resolution[:][-1] - 1
        if unique_dict_key(nob.dimension[:]) == '3d_xyz':
            out['extru_axis'] = 'z'
        else:
            out['extru_axis'] = 'axi'

    return {'generate': out}

def _periodize(nob):
    """method to setup periodization parameters"""
    pairs = []
    for pair in nob.bnd_patch[:]:
        pairs.append({'leader': pair["leader_patch"],
                      'follower': pair["follower_patch"]})
    out = pairs

    return {'periodize': out}

def _transform(nob):
    """method to setup dict form transform part"""
    out = {
        'rotate': {'angles': {'x': nob.rotation_angle_x[:],
                              'y': nob.rotation_angle_y[:],
                              'z': nob.rotation_angle_z[:]}},
        'translate': {'direction': nob.translation_vector[:]},
        'scale': {'scaling_factor': nob.scaling_vector[:]},
        'duplicate': {'nbr_sector_tot': nob.nbr_tot[:],
                      'nbr_sector_tgt': nob.nbr_tgt[:]}
    }

    return {'transform': out}


def _interpolate(nob):
    """method to setup read parameters"""
    if unique_dict_key(nob[:]) == 'interpolation_off':
        return {}
    else:
        out = {
            'files': [nob.src_mesh[:], nob.src_solution[:]],
            "meshtype": "hdf5",
            'patch_list': nob.patch_list_src[:]
            }

    return {'interpolate': out}

def _adapt(nob):
    """method to setup adapt parameters"""
    if unique_dict_key(nob[:]) == "adaptation_off":
        return {}
    else:
        out = {
            "factor":{
                "coefficient": nob.refinement_coeff[:],
                "hGrad": nob.hgrad[:],
                "hmin": nob.hmin[:]
            }
        }
        return {'adapt': out}

def _write(nob):
    """method to setup write parameters"""
    output_case = unique_dict_key(nob[:]).lower()

    if output_case == 'avbp hdf5':
        case_fmt = 'hdf5'
    elif output_case == 'cgns':
        raise NotImplementedError('CGNS format is not implemented yet for writing.')

    nob.geofile.set(nob.casename[:] + '.geo')
    out = {
        'casetype': case_fmt,
        'casename': nob.casename[:],
        'options': {},
    }
    if nob.find('options'):
        out['options'] = nob.options[:]

    return {'write': out}

if __name__ == "__main__":
    process_tab(process_finish)
